/*
 * Tabs are 4 symbols wide due to stupid Windows function names...
 * ---------------------------------------------------------------
 *
 * Copyright (c) 2010 KSI Labs, LLC. All rights reserved.
 * Written by Sergey Kubushyn <ksi@koi8.net>, <ksi@ksilabs.com>.
 *
 * File: hardware.c
 *
 * This is hardware-related functions for WDM driver for KSI Labs, LLC
 * CENTIPEDE PCI Board. These are generic functions, not Mach3 specific ones
 * that come in a separate file, mach3.c.
 *
 * Please read documentation or go to http://www.ksilabs.com for further info.
 */

/*
 * 						LICENSE and Terms of Use.
 *
 * First of all, this is NOT GPL or Free Software. It is open so you have
 * all the source code for the Engine. There is also VHDL source code for
 * CENTIPEDE board CPLD that is available from KSI Labs, LLC website,
 * http://www.ksilabs.com.
 *
 * As for this particular code, Windows Mach3 Engine, you are free to use it,
 * make any modifications to customize it for your own needs, fix inevitable
 * bugs, port it to other CNC software and do anything else as long as your
 * derivative work that is based on this code or uses parts or pieces thereof
 * is made to be used with any KSI Labs, LLC products.
 *
 * We do NOT grant you a permission to use any parts of this software in any
 * products that are designed for products not manufactured or distributed
 * by KSI Labs, LLC. We might give such a permission on case-by-case basis
 * but it is not legal to use this software or bits and pieces thereof with
 * products not coming from KSI Labs, LLC without our explicit written
 * permission.
 *
 * The original copyright notices must be included in all derivative works.
 * You do NOT have to ask our permission to make any modifications to this
 * code and you are not under obligation to give us any of your work back
 * as long as you are using it for your own needs.
 *
 * However we do NOT give our permission to distribute your derivative work
 * unless all original copyrights are retained, all the conditions of this
 * license are met (use only with KSI Labs, LLC products) and full source
 * code for such derivative work is available for free to everybody without
 * any further conditions including but not limited to signing any kinds of
 * NDA. We explicitely deny permission to distribute any derivative work
 * based on this code with closed source binary-only fragments or include
 * bits or pieces of this code in such products.
 *
 * CNC software like Mach3 and likes is excluded from these requirements as
 * long as it does not include parts or pieces of this software into the
 * close source proprietary application proper. It can use shared memory
 * buffer, IOCTL calls, etc. to communicate with this Engine or the one based
 * on it and keep all the software making those calls or working on that
 * buffer close source and proprietary as long as those calls/structures are
 * documented and that documentation is freely available to everybody without
 * any further conditions including but no limited to signing any kind of NDA.
 */

#include "precomp.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text (PAGE, AllocateDeviceResources)
#pragma alloc_text (PAGE, FreeDeviceResources)
#pragma alloc_text (PAGE, HWShutdown)
#endif


// ********************************************************************
// * This is supposed to be called from Power Management code for     *
// * changing our device power state. It should not be called at all  *
// * because we do not support Power Management. So this is just      *
// * a stub with possible debugging output to check if it was called. *
// ********************************************************************
NTSTATUS HWSetPower(PFDO_DATA FdoData, DEVICE_POWER_STATE DeviceState)
{
#ifdef DEBUG
	DebugPrint(TRACE, DBG_POWER, "<-->HWSetPower\n");
#endif
	return STATUS_SUCCESS;
}


// ***********************************************************************
// * This one shuts down our board before releasing all its resources    *
// * as a part of ReturnResources() processing on driver removal/unload. *
// * Return value doesn't have any sense; it is here because this one    *
// * gets called with KeSynchronizeExecution() the requires a BOOLEAN    *
// * return code.                                                        *
// ***********************************************************************
BOOLEAN HWShutdown(PFDO_DATA FdoData)
{
	UINT32	dummy;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_HW_ACCESS, "<-->HWShutdown\n");
#endif

	FdoData->IRQMask = 0;
	//
	// Reset the board. That will stop everything including IRQ generation.
	// All motion control will stop, all signals will go inactive, timer
	// will stop so no IRQs any more...
	//
	if (FdoData->CPDRegsBase) {
		FdoData->CPDRegsBase->mach_ctl = MACH_RESET;
		dummy = FdoData->CPDRegsBase->mach_ctl;
		dummy = FdoData->CPDRegsBase->irq_raw;	// Delay for MACH_RESET
		FdoData->CPDRegsBase->irq_mask = 0;
		dummy = FdoData->CPDRegsBase->irq_mask;
		FdoData->CPDRegsBase->mach_ctl = ESTOP_REQ;
		dummy = FdoData->CPDRegsBase->mach_ctl;
		FdoData->CPDRegsBase->irq_raw = 0xffffffff;
		dummy = FdoData->CPDRegsBase->irq_raw;
		FdoData->CPDRegsBase->dout = 0;
		dummy = FdoData->CPDRegsBase->dout;
	}

	FdoData->DrvState = DrvNotReady;
	FdoData->Running = FALSE;

	return TRUE;
}


// ***********************************************************************
// * Gets the HW resources assigned by the bus driver from the start-irp *
// * and maps it to system address space. Sets up the ISR. Also acquires *
// * all Parallel Ports that are available in the system.                *
// *                                                                     *
// * Only one BAR register is supported by CENTIPEDE-PCI board. This is  *
// * CPD_REGS Memory Mapped Base Address Register (BAR 0 at offset 10)   *
// ***********************************************************************
NTSTATUS AllocateDeviceResources(PFDO_DATA FdoData, PIRP Irp)
{
	PCM_PARTIAL_RESOURCE_DESCRIPTOR resourceTrans;
	PCM_PARTIAL_RESOURCE_LIST       partialResourceListTranslated;
	PIO_STACK_LOCATION              stack;
	ULONG                           i;
	NTSTATUS                        status = STATUS_SUCCESS;
	BOOLEAN 						bResInterrupt = FALSE;
	BOOLEAN							bResMemory = FALSE;
	ULONG                           numberOfBARs = 0;
	UINT32							dummy;

	PAGED_CODE();

#ifdef DEBUG
	DebugPrint(TRACE, DBG_INIT, "-->AllocateDeviceResources()\n");
#endif

	stack = IoGetCurrentIrpStackLocation(Irp);

	if (!stack->Parameters.StartDevice.AllocatedResourcesTranslated) {
		status = STATUS_DEVICE_CONFIGURATION_ERROR;
		goto End;
	}

	//
	// Parameters.StartDevice.AllocatedResourcesTranslated points
	// to a CM_RESOURCE_LIST describing the hardware resources that
	// the PnP Manager assigned to the device. This list contains
	// the resources in translated form. Use the translated resources
	// to connect the interrupt vector, map I/O space, and map memory.
	//

	partialResourceListTranslated = &stack->Parameters.StartDevice.\
					AllocatedResourcesTranslated->List[0].PartialResourceList;

	resourceTrans = &partialResourceListTranslated->PartialDescriptors[0];

	for (i = 0; i < partialResourceListTranslated->Count; i++, resourceTrans++) {

		switch (resourceTrans->Type) {

		case CmResourceTypePort:
#ifdef DEBUG
			DebugPrint(ERROR, DBG_INIT,
				"I/O mapped resource, should not happen for CENTIPEDE\n");
#endif
			status = STATUS_DEVICE_CONFIGURATION_ERROR;
			goto End;

		case CmResourceTypeMemory:

			numberOfBARs++;

			if (numberOfBARs == 1) {
#ifdef DEBUG
				DebugPrint(TRACE, DBG_INIT, "Memory mapped CSR:(0x%x:0x%x) Length:(%d)\n",
										resourceTrans->u.Memory.Start.LowPart,
										resourceTrans->u.Memory.Start.HighPart,
										resourceTrans->u.Memory.Length);
#endif
				//
				// Our PCI memory space should be 0x1000 in size.
				//
				ASSERT(resourceTrans->u.Memory.Length == 0x1000);
				FdoData->MemPhysAddress = resourceTrans->u.Memory.Start;
				FdoData->CPDRegsBase = MmMapIoSpace(
												resourceTrans->u.Memory.Start,
												CPD_MAP_IOSPACE_LENGTH,
												MmNonCached);
				if (!FdoData->CPDRegsBase) {
#ifdef DEBUG
					DebugPrint(ERROR, DBG_INIT, "MmMapIoSpace failed\n");
#endif
					status = STATUS_INSUFFICIENT_RESOURCES;
					goto End;
				}
#ifdef DEBUG
				DebugPrint(TRACE, DBG_INIT, "CPDRegsBase=0x%p\n", FdoData->CPDRegsBase);
#endif
				bResMemory = TRUE;

				break;
			}
#ifdef DEBUG
			DebugPrint(ERROR, DBG_INIT,
				"More than one BAR, shouldn't happen for CENTIPEDE\n");
#endif
			status = STATUS_DEVICE_CONFIGURATION_ERROR;
			goto End;

		case CmResourceTypeInterrupt:

			ASSERT(!bResInterrupt);

			bResInterrupt = TRUE;
			//
			// Save all the interrupt specific information in the device
			// extension because we will need it to disconnect and connect the
			// interrupt later on.
			//
			FdoData->InterruptLevel = (UCHAR)resourceTrans->u.Interrupt.Level;
			FdoData->InterruptVector = resourceTrans->u.Interrupt.Vector;
			FdoData->InterruptAffinity = resourceTrans->u.Interrupt.Affinity;

			if (resourceTrans->Flags & CM_RESOURCE_INTERRUPT_LATCHED) {

				FdoData->InterruptMode = Latched;

			} else {

				FdoData->InterruptMode = LevelSensitive;
			}

			//
			// Because this is a PCI device, we KNOW it must be
			// a LevelSensitive Interrupt.
			//

			ASSERT(FdoData->InterruptMode == LevelSensitive);
#ifdef DEBUG
			DebugPrint(TRACE, DBG_INIT,
				"Interrupt level: 0x%0x, Vector: 0x%0x, Affinity: 0x%x\n",
				FdoData->InterruptLevel,
				FdoData->InterruptVector,
				(UINT)FdoData->InterruptAffinity); // casting is done to keep WPP happy
#endif
			break;

		default:
#ifdef DEBUG
			DebugPrint(TRACE, DBG_INIT, "Unhandled resource type (0x%x)\n",
										resourceTrans->Type);
#endif
			break;
		}
	}

	//
	// Make sure we got all 2 resources to work with.
	//
	if (!(bResInterrupt && bResMemory)) {
		status = STATUS_DEVICE_CONFIGURATION_ERROR;
		goto End;
	}

	//
	// Reset and disable interrupts here before connecting interrupt
	//
	FdoData->CPDRegsBase->irq_mask = 0;
	dummy = FdoData->CPDRegsBase->irq_mask;
	FdoData->CPDRegsBase->irq_raw = 0xffffffff;
	dummy = FdoData->CPDRegsBase->irq_raw;
	FdoData->IRQMask = 0;

#ifdef _ISR_USE_DPC_
	IoInitializeDpcRequest(FdoData->Self, (PKDEFERRED_ROUTINE)CPDDpcForIsr);
#endif

	//
	// Register the interrupt
	//
	status = IoConnectInterrupt(&FdoData->Interrupt,
								CPDInterruptHandler,
								FdoData,
								NULL,
								FdoData->InterruptVector,
								FdoData->InterruptLevel,
								FdoData->InterruptLevel,
								FdoData->InterruptMode,
								TRUE,
								FdoData->InterruptAffinity,
								FALSE);

	if (status != STATUS_SUCCESS) {
#ifdef DEBUG
		DebugPrint(ERROR, DBG_INIT, "IoConnectInterrupt failed %x\n", status);
#endif
		goto End;
	}

	FdoData->Flags |= CPD_IRQ_IN_USE;

	//
	// Acquire all available Parallel Ports for our exclusive use.
	//
	FdoData->Port1isOK = FALSE;

	if (GetPPortInfo(FdoData, 0) != STATUS_SUCCESS) {
		RtlZeroMemory(&FdoData->PPortInfo[0], sizeof(PARALLEL_PORT_INFORMATION));
		goto End;
	}

	if (!FdoData->PPortInfo[0].TryAllocatePort(FdoData->PPortInfo[0].Context)) {
#ifdef DEBUG
		DebugPrint(ERROR, DBG_INIT, "Could not acquire PPort0\n");
#endif
		status = STATUS_INSUFFICIENT_RESOURCES;
		goto End;
	}

	FdoData->Port1isOK = TRUE;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_INIT, "PPort(0) Base = 0x%x\n",
						(ULONG)FdoData->PPortInfo[0].Controller);
	DebugPrint(TRACE, DBG_INIT, "PPort(0) Span = %i\n",
						(ULONG)FdoData->PPortInfo[0].SpanOfController);
#endif

	FdoData->Port2isOK = FALSE;

	if (GetPPortInfo(FdoData, 1) != STATUS_SUCCESS) {
		RtlZeroMemory(&FdoData->PPortInfo[1], sizeof(PARALLEL_PORT_INFORMATION));
		goto End;
	}

	if (!FdoData->PPortInfo[1].TryAllocatePort(FdoData->PPortInfo[1].Context)) {
#ifdef DEBUG
		DebugPrint(ERROR, DBG_INIT, "Could not acquire PPort1\n");
#endif
		status = STATUS_INSUFFICIENT_RESOURCES;
		goto End;
	}

	FdoData->Port2isOK = TRUE;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_INIT, "PPort(1) Base = 0x%x\n",
						(ULONG)FdoData->PPortInfo[1].Controller);
	DebugPrint(TRACE, DBG_INIT, "PPort(1) Span = %i\n",
						(ULONG)FdoData->PPortInfo[1].SpanOfController);
#endif

End:
#ifdef DEBUG
	DebugPrint(TRACE, DBG_HW_ACCESS, "<--AllocateDeviceResources()\n");
#endif

	return status;
}


// ***********************************************************************
// * Disconnect the interrupt and unmap all the memory and I/O resources *
// ***********************************************************************
NTSTATUS FreeDeviceResources(PFDO_DATA FdoData)
{
	UINT32	dummy;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_INIT, "-->FreeDeviceResources()\n");
#endif

	//
	// Disconnect interrupt
	//
	if ((FdoData->Flags & CPD_IRQ_IN_USE) != 0) {
		KeSynchronizeExecution(FdoData->Interrupt, HWShutdown, FdoData);
		IoDisconnectInterrupt(FdoData->Interrupt);
		FdoData->Interrupt = NULL;
		FdoData->Flags &= ~CPD_IRQ_IN_USE;
	}

	//
	// Unmap PCI Memory Space (CENTIPEDE registers)
	//
	if (FdoData->CPDRegsBase) {
		MmUnmapIoSpace((PVOID)FdoData->CPDRegsBase, CPD_MAP_IOSPACE_LENGTH);
		FdoData->CPDRegsBase = NULL;
	}

	//
	// Release Parallel Ports to the system.
	//
	if (FdoData->Port1isOK) {
		if (FdoData->PPortInfo[0].FreePort)
			FdoData->PPortInfo[0].FreePort(FdoData->PPortInfo[0].Context);
		FdoData->Port1isOK = FALSE;
	}

	if (FdoData->Port2isOK) {
		if (FdoData->PPortInfo[1].FreePort)
			FdoData->PPortInfo[1].FreePort(FdoData->PPortInfo[1].Context);
		FdoData->Port2isOK = FALSE;
	}

#ifdef DEBUG
	DebugPrint(TRACE, DBG_INIT, "<--FreeDeviceResources()\n");
#endif

	return STATUS_SUCCESS;
}

// *******************************************************************
// * Interrupt Handler, main routine. This is the real workhorse of  *
// * our Engine. It gets called on every CENTIPEDE timer tick that   *
// * is generated by the board when it is activated by the initial   *
// * IOCTL. It inspects the Mach3 structure on each and every tick   *
// * and takes appropriate actions. Input signals and board hardware *
// * indexes and various status information is transferred to the    *
// * Mach3 structure, output signals are transferred from that       *
// * structure to the board/parallel ports and to the actual output  *
// * pins, motion information from Mach3 is converted to proper      *
// * format and stuffed into board FIFO etc.                         *
// *                                                                 *
// * It also processes non-timer interrupts from CENTIPEDE and acts  *
// * appropriately.                                                  *
// *******************************************************************
BOOLEAN CPDInterruptHandler(PKINTERRUPT Interupt, PVOID ServiceContext)
{
	PFDO_DATA   			FdoData = (PFDO_DATA)ServiceContext;
	PMACH_STRUCT			mach = FdoData->MachSharedBlock;
	PCPD_REGS				brd = FdoData->CPDRegsBase;
    volatile PAXIS_DATA		AxData = FdoData->AxisData;
	UINT32					IrqStatus, BrdStatus, dummy, tmp;
	NTSTATUS				status;
	INT						i;

	//
	// If the adapter is in low power state, then it should not
	// recognize any interrupt. Should never happen--we do not
	// do any Power Management.
	//
	if (FdoData->DevicePowerState > PowerDeviceD0) return FALSE;

	//
	// Ditto if it is not in Running state
	//
	if (!FdoData->Running) return FALSE;

	//
	// Get Interrupt status
	//
	IrqStatus = brd->irq_stat;

	//
	// Skip if it is not our interrupt
	//
	if (IrqStatus == 0) return FALSE;

	//
	// Disable interrupts
	//
	brd->irq_mask = 0;

	//
	// Board status
	//
	BrdStatus = brd->mach_stat;


	///////////////////////////////////////////////////////
	// Now go check what IRQs we've got and process them //
	///////////////////////////////////////////////////////


	do {
		//
		// S@#t happened, indexes are no longer valid, everything's screwed.
		// Go into EStop, disable all but timer interrupts and do nothing
		// until all clear. Signal Emergency to Mach3, mark all Axes invalid,
		// all synchronizations lost, stop all motion but keep Running and
		// processing I/O signals on every timer interrupt.
		//
		if ((IrqStatus & CP_IRQ_MACH_FAULT) != 0) {

			MachStopSpindle(FdoData);

			PURGE_FIFO

			FdoData->DrvState = DrvEStopped;
			mach->EStop = TRUE;

			if ((BrdStatus & ST_LIMIT_HIT) != 0) {
				mach->EStopReason = LIMITSWITCH;
			} else if ((BrdStatus & ST_WD_FULL) != 0) {
				mach->EStopReason = DRIVERWATCHDOG;
			} else {
				mach->EStopReason = ESTOPINPUTLINE;
				mach->Emergency = TRUE;
			}

			for (i = 0; i < 6; i++) {
				if (mach->Busy) mach->Referenced[i] = FALSE;
				mach->Axis[i].Jogging = FALSE;
				mach->Axis[i].Probing = FALSE;
				mach->Axis[i].Homing = FALSE;
				mach->Axis[i].DeRef = FALSE;
				mach->Axis[i].AtSpeed = FALSE;
				mach->Axis[i].Acc = FALSE;
				mach->Axis[i].Dec = FALSE;
				mach->Axis[i].CurVelocity = 0;
				AxData[i].Active = FALSE;
				AxData[i].Hit = FALSE;
				AxData[i].QuantCntr = 0;
			}

			mach->Busy = FALSE;
			mach->Jogging = FALSE;

			//
			// Restore original PE Clock
			//
			if (!FdoData->ClockAtBase) {
				tmp = brd->mach_ctl;
				tmp &= ~MACH_CYCLE_MASK;
				tmp |= MACH_CYCLE_BITS(FdoData->BasePEClockDiv);
				brd->mach_ctl = tmp;
				tmp = brd->mach_ctl;
				FdoData->CurStatUnitsPerCycle = FdoData->BaseStatUnitsPerCycle;
			}

			FdoData->IRQMask = CP_IRQ_MACH_TMR; // | CP_IRQ_ADC | CP_IRQ_DAC;
			break;
		}

		//
		// Probe Switch hit. Call handler routine then disable the switch
		// (it is enabled when Probing pass is started,) acknowledge its
		// IRQ to unfreeze CENTIPEDE IDX registers so they will start
		// providing actual indexes, and remove that IRQ from IRQMask and
		// IrqStatus so it won't get reacked and reenabled on ISR exit.
		//
		if ((IrqStatus & CP_IRQ_MACH_PROBE) != 0) {
			//
			// Probe switch hit. Only process it if we are probing.
			// Skip ProbeHit proper for any other state (EStop it is.)
			// Ack IRQ, disable the switch, disable IRQ thus releasing
			// IDX registers from their "frozen" state.
			//
			if (FdoData->DrvState == DrvProbing)
				MachHandleProbeHit(FdoData);

			//
			// Disable ProbeHit IRQ and ProbeSwitch, reset ProbeHit IRQ.
			//
			FdoData->IRQMask &= ~CP_IRQ_MACH_PROBE;
			IrqStatus &= ~CP_IRQ_MACH_PROBE;

			tmp = brd->mach_cfg;
			tmp &= ~PROBE_SW_EN;
			brd->mach_cfg = tmp;
			dummy = brd->mach_cfg;

			brd->mach_stat = PROBE_SW_HIT;
			dummy = brd->mach_stat;

			brd->irq_raw = CP_IRQ_MACH_PROBE;
			dummy = brd->irq_raw;
		}

		//
		// One or more Home switches hit (we might be homing on several axes
		// at the same time.) Call handler routine then disable those
		// switches that were hit (they are enabled when Homing 1st Pass is
		// started,) acknowledge their IRQs to unfreeze corresponding IDX
		// registers so they will start giving actual indexes, and remove
		// those IRQs from IRQMask and IrqStatus so they won't get reacked
		// and reenabled on ISR exit.
		//
		if ((IrqStatus & CP_IRQ_ANY_HOME) != 0) {
			//
			// One of the Home switches hit. We only process this
			// if we are homing, otherwise ignore it.
			//
			if (FdoData->DrvState == DrvHoming1stPass)
				MachHandleHomeHit(FdoData);

			//
			// Disable HomeHit IRQ and HomeSwitch for those axes they were hit.
			// This is more complex than Probe IRQ because it is per axis.
			//
			tmp = brd->mach_cfg;
			if ((IrqStatus & CP_IRQ_MACH_HOME0) != 0) tmp &= ~AX_0_HOME_SW_EN;
			if ((IrqStatus & CP_IRQ_MACH_HOME1) != 0) tmp &= ~AX_1_HOME_SW_EN;
			if ((IrqStatus & CP_IRQ_MACH_HOME2) != 0) tmp &= ~AX_2_HOME_SW_EN;
			if ((IrqStatus & CP_IRQ_MACH_HOME3) != 0) tmp &= ~AX_3_HOME_SW_EN;
			if ((IrqStatus & CP_IRQ_MACH_HOME4) != 0) tmp &= ~AX_4_HOME_SW_EN;
			if ((IrqStatus & CP_IRQ_MACH_HOME5) != 0) tmp &= ~AX_5_HOME_SW_EN;
			brd->mach_cfg = tmp;
			dummy = brd->mach_cfg;

			brd->mach_stat = ST_HOME_HIT;
			dummy = brd->mach_stat;

			tmp = IrqStatus & CP_IRQ_ANY_HOME;
			IrqStatus &= ~tmp;
			FdoData->IRQMask &= ~tmp;

			brd->irq_raw = tmp;
			dummy = brd->irq_raw;
		}

		//
		// FIFO drained, running action is complete. Time to change state
		// depending on the action that just finished. We reset DONE IRQ
		// in IRQMask so it won't be set after interrupts are acked and
		// reenabled on ISR exit. It will be set as needed when some actions
		// are done and want to wait until FIFO is drained.
		//
		if ((IrqStatus & CP_IRQ_MACH_DONE) != 0) {
			//
			// No breaks...
			//
			switch (FdoData->DrvState) {
				case DrvJogFinishing:
				case DrvProgramFinishing:
					for (i = 0; i < 6; i++) {
						AxData[i].Active = FALSE;
						AxData[i].Hit = FALSE;
						mach->Axis[i].Jogging = FALSE;
						//mach->Axis[i].Acc = FALSE;
						mach->Axis[i].Dec = FALSE;
					}

					RESTORE_LIMITS

					if (mach->Homing) MachCompleteHome(FdoData);
					
					mach->Jogging = FALSE;

					if (FdoData->DrvState != DrvProgramFinishing)
						mach->Sync = TRUE;

				case DrvWaitingForHome:
				case DrvWaitingForDeref:
				case DrvWaitingForClear:
					for (i = 0; i < 6; i++) {
						AxData[i].QuantCntr = 0;
						mach->Axis[i].CurVelocity = 0;
					}

					FdoData->StateChange = TRUE;

				default:
					FdoData->IRQMask &= ~CP_IRQ_MACH_DONE;
			}
		}

#if 0
		//
		// ADC has finished conversion, got data, and ready for the next cycle
		//
		if ((IrqStatus & CP_IRQ_ADC) != 0) {

			switch (FdoData->LastADC) {
				case 0:
					brd->adc = ADC_IN1;
					FdoData->ADCData[3] = brd->adc;
					FdoData->LastADC = 1;
					break;
				case 1:
					brd->adc = ADC_IN2;
					FdoData->ADCData[0] = brd->adc;
					FdoData->LastADC = 2;
					break;
				case 2:
					brd->adc = ADC_IN3;
					FdoData->ADCData[1] = brd->adc;
					FdoData->LastADC = 3;
					break;
				case 3:
					brd->adc = ADC_IN0;
					FdoData->ADCData[2] = brd->adc;
					FdoData->LastADC = 0;
					break;
				default:
					//
					// Dummy cycle after init, data is garbage so it is
					// thrown away and Channel 0 cycle initiated.
					//
					brd->adc = ADC_IN0;
					tmp = brd->adc;
					FdoData->LastADC = 0;
					break;
			}
		}

		//
		// DAC SPI finished sending data and ready for the next cycle
		//
		if ((IrqStatus & CP_IRQ_DAC) != 0) {

			switch (FdoData->LastDAC) {
				case 0:
					brd->dac = DAC_B_WORD(FdoData->DACData[1]);
					FdoData->LastDAC = 1;
					break;
				case 1:
					brd->dac = DAC_A_WORD(FdoData->DACData[0]);
					FdoData->LastDAC = 0;
					break;
				default:
					//
					// Should not happen but just in case -- do DAC_B cycle
					//
					brd->dac = DAC_B_WORD(FdoData->DACData[1]);
					FdoData->LastDAC = 1;
					break;
			}
			tmp = brd->irq_raw;
		}
#endif

		//
		// Spindle Timing Pulse received. That comes from multi-pulse
		// per revolution sensor and used for measuring Spindle RPM if
		// the machine is equipped with such a sensor. Otherwise the
		// single-pulse per revolution Index sensor is used for both
		// Threading start and RPM measurement.
		//
		if ((IrqStatus & CP_SPINDLE_TIMING) != 0) {
			//
			// We are adding each count between timing pulses to
			// PartSpindleCount. The final value is moved to PPRSpindleCount
			// when Index pulse comes and PartSpindleCount is zeroed so
			// we start over on each Index pulse. If no Timing Sensor
			// installed we use a single count between Index pulses that is
			// taken on Index IRQ.
			//
			// Actual RPM calculation and reporting it to Mach is done on
			// each Index IRQ.
			//
			// However Pulse Engine Clock is updated here when threading.
			// It can go very high rate, something like 85 KHz at 5,000 RPM
			// when 1024 PPR sensor is used. That means we should avoid using
			// such high PPR sensors if that high RPM is planned. Updating
			// PE Clock here gives us best tracking possible. If there is no
			// Timing Sensor installed PE Clock is updated on each Index IRQ
			// i.e. once per revolution. That results in worse tracking but
			// might be adequate if our Spindle has big enough inertia.
			//
			// We do also count Timing IRQs here to find out how many PPR
			// our sensor generates that is necessary for calculating
			// PE Clock updates when threading (calculations are done against
			// per-revolution value.)
			//
			// We do NOT check if Timing Sensor is enabled here because the
			// very fact we got here means it is.
			//
			FdoData->nSlotsCount++;

			FdoData->SpindleCount = brd->mach_sp_cnt;
			FdoData->PartSpindleCount += FdoData->SpindleCount;
			//
			// Now adjust Pulse Engine Clock if Threading.
			//
			if (mach->Threading) MachUpdatePEClock(FdoData);
#ifdef DEBUG
		DebugPrint(TRACE, DBG_INTERRUPT, "\n-###- TIMING%s: %d\n", mach->Threading ? "(Threading)" : "", FdoData->SpindleCount);
#endif
		}

		//
		// Spindle Index Pulse received.
		//
		if ((IrqStatus & CP_SPINDLE_IDX) != 0) {
			//
			// We were waiting for Index pulse to cock the threading trigger
			// and start filling FIFO. We signal a StateChange so it will
			// be processed by the FSM further down on next Timer IRQ.
			//
			if ((FdoData->DrvState == DrvPrepForThreading) &&
							(mach->TrajHead != mach->TrajIndex)) {
				FdoData->StateChange = TRUE;
			}

			if (!mach->InSigs[TIMING].Active) {
				//
				// No Timing Sensor, we are the last resort. Get counts
				// and update PE Clock if threading.
				//
				FdoData->PPRSpindleCount = brd->mach_sp_cnt;
				FdoData->SpindleCount = FdoData->PPRSpindleCount;
				mach->nSlots = 0;
#ifdef DEBUG
		DebugPrint(TRACE, DBG_INTERRUPT, "\n-###- INDEX%s: %d\n", mach->Threading ? "(Threading)" : "", FdoData->SpindleCount);
#endif
				if (mach->Threading) MachUpdatePEClock(FdoData);
			} else {
				//
				// There IS Timing Sensor and we are here because a full
				// revolution is complete. We take the accumulated Spindle
				// count and update nSlots with what we have counted. No
				// PE Clock update here; it is done on Timing Sensor IRQ.
				//
				// PartSpindleCount and nSlotsCount are zeroed for the next
				// revolution.
				//
				mach->nSlots = FdoData->nSlotsCount;
				FdoData->nSlotsCount = 0;
				FdoData->PPRSpindleCount = FdoData->PartSpindleCount;
				FdoData->PartSpindleCount = 0;
#ifdef DEBUG
		DebugPrint(TRACE, DBG_INTERRUPT, "\n-###- INDEX/TIMING%s: %d\n", mach->Threading ? "(Threading)" : "", FdoData->PPRSpindleCount);
#endif
			}

			//
			// Update Mach3 RPM DRO. It is done here because it makes no sense
			// to do it more often -- no data changed until next revolution
			// is complete. If Spindle is stopped the proper update will be
			// done in Timer IRQ (there will be no Index IRQs so we won't get
			// here to make an update.) This is also used for calculating
			// "TrueSpindleRPM" in Mach3 for planning. We lock this for the
			// duration of a threading pass so Mach3 plans for a constant RPM
			// and we compensate for RPM changes by comparing this value to
			// actual measured RPM.
			//
			// RPM DRO reporting is kinda weird but that is how it is done
			// in Mach3...
			//
			if (!mach->Threading) {
				FdoData->ThrSpindlePPR = FdoData->PPRSpindleCount;
				((PINT)&mach->CircData[0])[12] = FdoData->PCIFreq /
													FdoData->BasePEClockDiv;
				((PINT)&mach->CircData[0])[5] = FdoData->PPRSpindleCount /
													FdoData->BasePEClockDiv;
			}
		}

		////////////////
		// Timer Tick //
		////////////////

		if ((IrqStatus & CP_IRQ_MACH_TMR) != 0) {
			//
			// Propagate I/O signals from/to Mach3, update indexes, Spindle,
			// encoders, MPGs, ADC/DAC, etc...
			//
			MachGetInputs(FdoData);
			MachCheckTHC(FdoData);
			MachUpdateIndexes(FdoData);
			MachUpdateADC(FdoData);
			MachUpdateDataIn(FdoData);
			MachUpdateDataOut(FdoData);
			MachPutOutputs(FdoData);
			MachUpdateDAC(FdoData);
			//
			// Spindle stopped, update.
			//
			if ((BrdStatus & SPINDLE_STOPPED) != 0) {
				FdoData->ThrSpindlePPR = FdoData->PPRSpindleCount = 0;
				FdoData->SpindleCount = FdoData->PartSpindleCount = 0;
				((PINT)&mach->CircData[0])[12] = FdoData->PCIFreq /
												FdoData->BasePEClockDiv;
				((PINT)&mach->CircData[0])[5] = FdoData->PPRSpindleCount /
												FdoData->BasePEClockDiv;
			}
			MachUpdateSpindle(FdoData);
			// This is MachUpdateEncoders() :)
			if ((tmp = brd->mach_enc) != 0) {
				for (i = 0; i < 4; i++) {
					if (!mach->Encoders[i].Active) continue;
					mach->Encoders[i].Count[0] += (CHAR)((tmp >> (i*8)) & 0xff);
				}
			}
			MachUpdateMPGs(FdoData);

			RESTORE_LIMITS

			//
			// EStop requested by Mach3, go into EStop unconditionally.
			//
			if (mach->EStop && FdoData->DrvState != DrvEStopped) {

				MachStopSpindle(FdoData);

				PURGE_FIFO

				//
				// MACH_DONE IRQ may be raised when FIFO purged, clear it.
				//
				brd->irq_raw = CP_IRQ_MACH_DONE;
				dummy = brd->irq_raw;

				FdoData->DrvState = DrvEStopped;

				tmp = brd->mach_ctl;
				tmp |= ESTOP_REQ;
				tmp &= ~MACH_CYCLE_MASK;
				tmp |= MACH_CYCLE_BITS(FdoData->BasePEClockDiv);
				brd->mach_ctl = tmp;
				dummy = brd->mach_ctl;
				FdoData->CurStatUnitsPerCycle = FdoData->BaseStatUnitsPerCycle;
				FdoData->ClockAtBase = TRUE;
				FdoData->IRQMask = CP_IRQ_MACH_TMR | CP_IRQ_MACH_FAULT; //| CP_IRQ_ADC | CP_IRQ_DAC;
				for (i = 0; i < 6; i++) {
					AxData[i].Active = FALSE;
					AxData[i].Hit = FALSE;
					AxData[i].QuantCntr = 0;
					if (mach->Busy) mach->Referenced[i] = FALSE;
					mach->Axis[i].Jogging = FALSE;
					mach->Axis[i].Probing = FALSE;
					mach->Axis[i].Homing = FALSE;
					mach->Axis[i].DeRef = FALSE;
					mach->Axis[i].AtSpeed = FALSE;
					//mach->Axis[i].Acc = FALSE;
					mach->Axis[i].Dec = FALSE;
					mach->Axis[i].CurVelocity = 0;
				}

				mach->Busy = FALSE;
				mach->Jogging = FALSE;

				break;
			}

			//
			// Mach3 attempts to go out of EStop. It resets EStop.
			// MachLeaveEStop() is invoked and if it is successfull we are
			// going into Idle state and setting StateChange so the driver
			// will be able to go into any state requested by Mach3.
			//
			// We stay in EStop and break out of the loop without any further
			// processing if there is no attempt to go out by Mach3 or
			// MachLeaveEStop() failed. The latter means we still have
			// condition that forced us into EStop (e.g. BigRedButton is still
			// depressed or we are on a Limit Switch.)
			//
			if (FdoData->DrvState == DrvEStopped) {

				if (mach->EStop) break;

				if ((mach->EStopReason = MachLeaveEStop(FdoData)) != STATUS_SUCCESS) {
					mach->EStop = TRUE;
					break;
				}

				FdoData->DrvState = DrvIdle;
				mach->Busy = FALSE;
				mach->Jogging = FALSE;
				FdoData->IRQMask = CP_IRQ_MACH_TMR | CP_IRQ_MACH_FAULT; // | CP_IRQ_ADC | CP_IRQ_DAC;
				if (mach->InSigs[INDEX].Active) FdoData->IRQMask |= CP_SPINDLE_IDX;
				if (mach->InSigs[TIMING].Active) FdoData->IRQMask |= CP_SPINDLE_TIMING;
				IrqStatus |= FdoData->IRQMask;
				FdoData->StateChange = TRUE;
				break;
			}

			//
			// We are waiting for FIFO to drain. Do nothing; break out that
			// will acknowledge and retrigger interrupts. We do this only if
			// StateChange is not set (that means we ARE waiting.) If it IS
			// set we've got what we were waiting for so it is time to proceed
			// with that state change; the wait is over. Another reason to
			// wait is Threading -- Mach3 is preparing the buffer and we are
			// waiting until it sets WaitForTrigger.
			//
			if (!FdoData->StateChange) {

				if (FdoData->DrvState == DrvJogFinishing) break;
				if (FdoData->DrvState == DrvWaitingForHome) break;
				if (FdoData->DrvState == DrvWaitingForDeref) break;
				if (FdoData->DrvState == DrvWaitingForClear) break;
				if (FdoData->DrvState == DrvPrepForThreading) break;
				if (FdoData->DrvState == DrvProgramFinishing) break;

				//					--------
				//					Dwelling
				//					--------
				// No motion, just decrementing DwellTime down to zero the
				// change state to whatever requested by Mach3. We can only
				// dwell between other states (Dwelling is a state itself)
				// so we switch to Idle after Dwell is over. It can not
				// happen inside a multi-step state (e.g. Homing) so we are
				// OK with going Idle after Dwell.
				//
				// N.B. We are subtracting 5 here because Dwell time is in
				//      1/KernelFreq units and we have a timer interrupt
				//      every 5 such units.
				//
				// N.B. We keep DwellTime at 1 (not going to zero) if
				//		SpindleStopping is set thus preventing Dwell from
				//		expiring before Step/Dir Spindle is fully stopped.
				//		This is done to prevent Mach from starting it in
				//		reverse before it fully stopped.
				//
				if (FdoData->DrvState == DrvDwelling) {
					if ((mach->DwellTime -= 5) <= 0) {
						if (!FdoData->SpindleStopping) {
							mach->DwellTime = 0;
							FdoData->DrvState = DrvIdle;
							FdoData->StateChange = TRUE;
						} else {
							mach->DwellTime = 1;
						}
					}
					break;
				}

				//					--------------
				//					Planned motion
				//					--------------
				// We are still in that state if nothing bad happened (e.g.
				// limit switch hit) and TrajIndex != TrajHead. It is done
				// when TrajIndex == TrajHead so we go to Finishing
				// state and start waiting for DONE IRQ (FIFO drained.)
				// We are also doing the same planned motion when we are
				// Threading after SpindleIDX has been signalled for the
				// first time when we were in DrvWaitingForThrIDX state.
				// That would release the hold and we will go into Threading
				// mode immediately. Threading is exactly the same as
				// Planned Motion with only difference that Pulse rate is
				// constantly adjusted in MachUpdatePEClock() to match
				// current Spindle RPM.
				//
				if (FdoData->DrvState == DrvInProgram) {
					if (mach->Pierce > 0) {
						mach->Pierce = (mach->Pierce >= 5) ? mach->Pierce - 5 : 0;
					} else if ((mach->TrajIndex & 0xfff) != (mach->TrajHead & 0xfff)) {
						status = MachPlannedMotion(FdoData);
						if (status != STATUS_SUCCESS) {
							FdoData->DrvState = DrvEStopped;
							mach->EStop = TRUE;
							//
							// We do NOT have "Limit Switch Hit" error here.
							// It will raise a separate interrupt if hit and
							// it will be processed separately.
							//
							if (status == STATUS_DEVICE_NOT_READY)
								mach->EStopReason = CONFIG;
							else
								mach->EStopReason = SOFTLIMITS;
						}

					} else {
						//
						// No more job, wait for FIFO to drain.
						//
						FdoData->DrvState = DrvProgramFinishing;
						FdoData->IRQMask |= CP_IRQ_MACH_DONE;
					}
					break;
				}

				//					-------
				// 					Jogging
				//					-------
				// MachJog() returns STATUS_SUCCESS when it is done with
				// stuffing all data in the FIFO and there is nothing more
				// to do. It returns STATUS_MORE_PROCESSING_REQUIRED if it
				// is not done i.e. we are still jogging.
				//
				if (FdoData->DrvState == DrvJogging) {
					status = MachJog(FdoData);
					if (status == STATUS_SUCCESS) {
						FdoData->DrvState = DrvJogFinishing;
						FdoData->IRQMask |= CP_IRQ_MACH_DONE;
					}
					break;
				}

				//					-------
				// 					Probing
				//					-------
				// MachJog() returns STATUS_SUCCESS when it is done with
				// stuffing all data in FIFO and there is nothing more to do.
				// It returns STATUS_MORE_PROCESSING_REQUIRED if it is not
				// done. Jog set for Deceleration when either Probe switch
				// hit or TripPoint is reached. When Probe switch hit the hit
				// position is frozen in CENTIPEDE IDX registers and PROBE_HIT
				// IRQ is generated that is processed above. That processing
				// unfreezes IDX registers and they keep updating with current
				// position (hit point freeze does NOT make CENTIPEDE loose
				// IDX--it keeps updating the counters behind the transparent
				// latches that IDX registers are.) All axes
				// Probing/Jogging/Homing variables are set by Mach3 and
				// reset by MachJog() so there is no need to check and
				// manipulate them here.
				//
				if (FdoData->DrvState == DrvProbing) {
					status = MachJog(FdoData);
					if (status == STATUS_SUCCESS) {
						FdoData->DrvState = DrvJogFinishing;
						FdoData->IRQMask |= CP_IRQ_MACH_DONE;
					}
					break;
				}

				//					------
				// 					Homing
				//					------
				// This is the most complex of all jogs. Homing is done in
				// two steps. First pass is to hit all Home switches on
				// axes we are homing. For enslaved configurations (gantry
				// style routers etc. with separate drives/axes) we continue
				// to jog on Master/Slave pair until BOTH switches hit and
				// only then decelerate to stop. Chances are they are NOT hit
				// at the same position i.e. gantry is out of square. Both
				// axes Home switch hit positions are stored in AxisData
				// structures.
				// The second pass is Dereferencing where axes are moved back
				// just off of Home switches. This also uses MachJog() but it
				// behaves differently -- Master and Slave axes are NOT locked
				// to each other if there is a difference in Home switch hit
				// position so they are moved different distances to make them
				// stop at the same position thus squaring them.
				// Dereference move is calculated in MachHandleHomeHit()
				// routine. It also sets up all the required Axis[x]
				// variables for that move.
				//
				// But wait, there is more :) There is yet another pass before
				// the first one. Mach3 will setup a jog in opposite direction
				// if an axis is ALREADY on the home switch. The idea is that
				// we usually move TOWARDS the switch until it's hit but if we
				// are already on that switch we should move in the opposite
				// direction until the switch is RELEASED.
				//
				// That might seem OK but actually it is a bad idea. First of
				// all switches do NOT usually open and close at the very same
				// point, there is hysteresis (backlash :)) That's why it is
				// better to always approach them from the same direction.
				// Then it might cause troubles on a slaved axes because if
				// they are out of square they might be at such a position
				// that only one switch is activated so Mach3 will command
				// them to move in opposite directions. That would've been OK
				// (less switch backlash) but the algorithm for enslaved pair
				// is that they move together in the same direction until BOTH
				// switches hit then they back up independently to stop on
				// the hit point thus squaring the gantry. That means that
				// we won't hit one of switches no matter which way we move.
				//
				// That's why there is an additional pass before we start. It
				// checks if any home switch is activated for axes we are
				// about to home and moves back all the affected axes until
				// ALL switches are released. Then we run a regular sequence
				// with affected Axis[x].JoggDir inverted (they were setup to
				// move OFF the switch, not TOWARDS it by Mach3.)
				//
				// There is just another quirk -- some setups don't have
				// separate Home and Limit switches; they use one of the Limit
				// switches as Home by wiring both CENTIPEDE terminals to the
				// same switch. Such configuration is NOT recommended and
				// should NOT be used but anyways here it is...
				//
				// For such a configuration to work one has to check one bogus
				// checkbox, "Pins 2-9 as inputs" for Port 2 in "Ports & Pins"
				// configuration dialog. This will DISABLE ALL LIMIT switches
				// functionality for the duration of entire Homing sequence.
				// Other Port 2 settings do nothing, they are not used for
				// anything in CENTIPEDE Driver.
				//
				if (FdoData->DrvState == DrvHomingPrePass) {
					status = MachJog(FdoData);
					if (status == STATUS_SUCCESS) {
						FdoData->DrvState = DrvWaitingForHome;
						FdoData->IRQMask |= CP_IRQ_MACH_DONE;
					}
					break;
				}

				if (FdoData->DrvState == DrvHoming1stPass) {
					status = MachJog(FdoData);
					if (status == STATUS_SUCCESS) {
						FdoData->DrvState = DrvWaitingForDeref;
						FdoData->IRQMask |= CP_IRQ_MACH_DONE;
					}
					break;
				}

				if (FdoData->DrvState == DrvHomingDeref) {
					status = MachJog(FdoData);
					if (status == STATUS_SUCCESS) {
						FdoData->DrvState = DrvWaitingForClear;
						FdoData->IRQMask |= CP_IRQ_MACH_DONE;
					}
					break;
				}

				if (FdoData->DrvState == DrvHomingClearSwitches) {
					status = MachJog(FdoData);
					if (status == STATUS_SUCCESS) {
						FdoData->DrvState = DrvJogFinishing;
						FdoData->IRQMask |= CP_IRQ_MACH_DONE;
					}
					break;
				}

			} else {	// Some state ended; find out what to do next

				//
				// Wait for Mach3 to synchronize to new indexes
				//
				if (mach->Sync) break;

				FdoData->StateChange = FALSE;

				//
				// Check if we are simulating or not, set mach_ctl.
				//
				tmp = brd->mach_ctl;
				if (mach->Simulate)
					tmp |= SIMULATE;
				else
					tmp &= ~SIMULATE;
				brd->mach_ctl = tmp;
				dummy = brd->mach_ctl;

				//
				// We might've finished a threading pass. Restore
				// Pulse Engine Clock to base value. It might be not
				// needed (we were not threading) but won't hurt.
				//
				if (!mach->Threading && !FdoData->ClockAtBase) {
					tmp = brd->mach_ctl;
					tmp &= ~MACH_CYCLE_MASK;
					tmp |= MACH_CYCLE_BITS(FdoData->BasePEClockDiv);
					brd->mach_ctl = tmp;
					tmp = brd->mach_ctl;
					FdoData->CurStatUnitsPerCycle = FdoData->BaseStatUnitsPerCycle;
					FdoData->ClockAtBase = TRUE;
#ifdef DEBUG
		DebugPrint(TRACE, DBG_INTERRUPT, "\n-###- Restore Base Clock\n");
#endif
				}

				//
				// Dwelling...
				//
				if (mach->DwellTime > 0) {
					FdoData->DrvState = DrvDwelling;
					break;
				}

				//
				// Homing, preparation pass (move off home switches
				// if activated and setup proper jog directions for
				// subsequent homing passes) finished. Go to actual
				// homing 1st pass.
				//
				if (FdoData->DrvState == DrvWaitingForHome) {
					if (mach->Homing) {
						MachPrepareHome1stPass(FdoData);
						FdoData->DrvState = DrvHoming1stPass;
						mach->Busy = TRUE;
						mach->Jogging = TRUE;
						tmp = dummy = 0;
						for (i = 0; i < 6; i++) {
							//
							// Enable home switches and their IRQs for axes
							// we are homing on. Also set Jogging on them that
							// might've been reset by PrePass.
							//
							if (mach->Axis[i].Homing) {
								tmp |= 1 << ((i*5)+4);
								dummy |= 1 << (i+7);
							}
						}
						brd->mach_stat = ST_HOME_HIT;
						brd->mach_cfg |= tmp;
						tmp = brd->mach_cfg;
						brd->irq_raw = dummy;
						tmp = brd->irq_raw;
						FdoData->IRQMask |= dummy;
					} else {
						MachAbortHome(FdoData);
						IrqStatus |= FdoData->IRQMask;
					}
					break;
				}

				//
				// Homing, 1st pass finished, run second pass
				//
				if (FdoData->DrvState == DrvWaitingForDeref) {
					if (mach->Homing) {
						MachPrepareDeref(FdoData);
						FdoData->DrvState = DrvHomingDeref;
						mach->Busy = TRUE;
						mach->Jogging = TRUE;
					} else {
						MachAbortHome(FdoData);
						IrqStatus |= FdoData->IRQMask;
					}
					break;
				}

				//
				// Last step of Homing -- moving off of switches.
				//
				if (FdoData->DrvState == DrvWaitingForClear) {
					if (mach->Homing) {
						if (MachPrepareClearSwitches(FdoData) != STATUS_SUCCESS) {

							FdoData->DrvState = DrvHomingClearSwitches;
							mach->Busy = TRUE;
							mach->Jogging = TRUE;

						} else {

							FdoData->DrvState = DrvIdle;
							RESTORE_LIMITS
							MachCompleteHome(FdoData);
							mach->Busy = FALSE;
							mach->Jogging = FALSE;
							mach->Sync = TRUE;
							FdoData->StateChange = TRUE;
						}

					} else {
						MachAbortHome(FdoData);
						IrqStatus |= FdoData->IRQMask;
					}

					break;
				}

				//					-------------------
				//					Threading, 2nd step
				//					-------------------
				// Now everything is almost ready, Mach3 put some steps
				// in motion buffer, we are waiting for SpindleIDX. Once it
				// is detected we cock the trigger in CENTIPEDE so it will
				// start planned motion cycle that very moment SpindleIDX
				// has been detected on the next revolution.
				// We fill FIFO with first motion entries so there will be
				// data to output. The entire sequence is pure hardware --
				// CENTIPEDE Pulsing Engine will hold all motion until
				// SpindleIDX came then it will start putting steps as if it
				// is regular planned motion. Pulsing Engine state is reset
				// by the hardware when SpindleIDX comes so the timing will
				// be right starting with the very first step.
				// Pulsing Engine base clock will be adjusted every Spindle
				// revolution (or every Timing pulse if there is a timing
				// sensor with multiple pulses per revolution installed) by
				// MachUpdatePEClock() to match actual RPM for proper per-rev
				// feed rate. That will be done for as long as mach->Threading
				// is set then it will return to its original value set by
				// SetTimerFreq IOCTL when Mach3 was started.
				//
				if (FdoData->DrvState == DrvPrepForThreading) {
					mach->WaitForTrigger = FALSE;
////					mach->Threading = TRUE;
					mach->Busy = TRUE;
					tmp = brd->mach_ctl;
					tmp |= WAIT4PULSE;
					brd->mach_ctl = tmp;
					tmp = brd->mach_ctl;
					MachUpdatePEClock(FdoData);
					FdoData->DrvState = DrvIdle;
					FdoData->StateChange = TRUE;
#ifdef DEBUG
		DebugPrint(TRACE, DBG_INTERRUPT, "\n-###- THREADING 2nd Step\n");
#endif
					break;
				}


				//					-------------------
				//					Threading, 1st step
				//					-------------------
				// We do nothing yet, just switch to DrvPrepForThreading
				// state, no StateChange is set. In Timer IRQ with no
				// StateChange we will be just breaking out of the IRQ loop
				// waiting for Spindle IDX IRQ. That is done to make sure
				// we won't change state just before that IRQ but just
				// AFTER it. That will ensure we have ample time to lock
				// the Pulse Engine into WaitForPulse mode and fill out
				// FIFO before next Spindle IDX IRQ came (they are just
				// 100Hz even for 6,000 RPM.)
				//
				if (mach->WaitForTrigger) {
//					mach->WaitForTrigger = FALSE;
//					mach->Threading = TRUE;
					mach->Busy = FALSE;
					FdoData->DrvState = DrvPrepForThreading;
#ifdef DEBUG
		DebugPrint(TRACE, DBG_INTERRUPT, "\n-###- THREADING 1st Step\n");
#endif
					break;
				}

				//
				// Check if we have some planned motion to do
				//
				if ((mach->TrajIndex & 0xfff) != (mach->TrajHead & 0xfff)) {
					for (i = 0; i < 6; i++) {
						AxData[i].PendingIDX = mach->Axis[i].Index;
						AxData[i].Active = FALSE;
						AxData[i].Hit = FALSE;
						AxData[i].QuantCntr = 0;
					}
					mach->Busy = TRUE;
					mach->Jogging = FALSE;
					FdoData->DrvState = DrvInProgram;
					break;
				}

				//
				// Now check for various jogs -- probing, homing, jogging...
				//
				for (i = 0; i < 6; i++) {

					if (!mach->Axis[i].Jogging) continue;

					if (mach->Homing) {
						FdoData->DrvState = DrvHomingPrePass;
						break;
					}

					if (mach->Axis[i].Probing) {
						FdoData->DrvState = DrvProbing;
						break;
					}

					FdoData->DrvState = DrvJogging;
				}

				//
				// First step of Homing -- PrePass. It is only entered if
				// HomeSwitchSecurity is off in Mach3 config and one of the
				// Home switches is activated on an Axis (or Master/Slave
				// pair, at least one switch in this case.) Mach3 will set
				// those axes for a DeRef move (DeRef and Jogging set,
				// JoggDir is set to move OFF the switches, Homing is NOT
				// set for those axes but global Homing IS set.)
				//
				// We do not work Mach3 way. We will set Homing on those
				// axes, reset DeRef, set Master/Slave pairs to move in the
				// same direction OFF the switches and then jog until the
				// corresponding switches are cleared. Then we will do
				// regular Homing cycle.
				//
				// ClearSwitchesMask is here to save some instructions in Jogging
				// routine -- we just use it there to check Home switches
				// state instead of recalculating it every timer tick. Not
				// probably a big gain but MachJog routine is called every
				// timer tick when jogging while this state switch is only
				// called once in a very long while...
				//
				// This one is skipped if we start Homing with switches OFF.
				//
				if (FdoData->DrvState == DrvHomingPrePass) {

					tmp = brd->din;
					tmp ^= FdoData->HomeSWInvMask;
					tmp &= FdoData->HomeSWEnMask;

					if ((tmp & ST_HOME_HIT) == 0) {

						FdoData->DrvState = DrvWaitingForHome;
						FdoData->StateChange = TRUE;

					} else {

						for (i = 0; i < 6; i++) {

							AxData[i].PendingIDX = mach->Axis[i].Index;
							AxData[i].Active = FALSE;
							AxData[i].QuantCntr = 0;
							mach->Axis[i].CurVelocity = 0;

							if (AxData[i].Hit) continue;

							AxData[i].ClearSwitchesMask = 0;

							if (mach->Axis[i].DeRef) {

								AxData[i].Hit = TRUE;
								mach->Axis[i].DeRef = FALSE;
								mach->Axis[i].Homing = TRUE;
								mach->Axis[i].Jogging = TRUE;

								AxData[i].ClearSwitchesMask |= 1 << (i*3+2);

								if (AxData[i].Type == Single) continue;

								if (AxData[i].Type == Master) {

									AxData[AxData[i].Slave].Hit = TRUE;
									mach->Axis[AxData[i].Slave].DeRef = FALSE;
									mach->Axis[AxData[i].Slave].Homing = TRUE;
									mach->Axis[AxData[i].Slave].Jogging = TRUE;
									mach->Axis[AxData[i].Slave].JoggDir =
													mach->Axis[i].JoggDir;
									AxData[i].ClearSwitchesMask |=
													1 << (AxData[i].Slave*3+2);
									AxData[AxData[i].Slave].ClearSwitchesMask =
													AxData[i].ClearSwitchesMask;

								} else {

									AxData[AxData[i].Master].Hit = TRUE;
									mach->Axis[AxData[i].Master].DeRef = FALSE;
									mach->Axis[AxData[i].Master].Homing = TRUE;
									mach->Axis[AxData[i].Master].Jogging = TRUE;
									mach->Axis[AxData[i].Master].JoggDir =
														mach->Axis[i].JoggDir;
									AxData[i].ClearSwitchesMask |=
													1 << (AxData[i].Master*3+2);
									AxData[AxData[i].Master].ClearSwitchesMask =
													AxData[i].ClearSwitchesMask;
								}

							} else {

								AxData[i].Hit = FALSE;
								mach->Axis[i].Jogging = FALSE;
							}
						}
					}

					mach->Busy = TRUE;
					mach->Jogging = TRUE;

					RESTORE_LIMITS

					break;
				}

				if (FdoData->DrvState == DrvProbing) {

					mach->Busy = TRUE;
					mach->Jogging = TRUE;
					for (i = 0; i < 6; i++) {
						AxData[i].PendingIDX = mach->Axis[i].Index;
						AxData[i].Active = FALSE;
						AxData[i].Hit = FALSE;
						AxData[i].QuantCntr = 0;
						mach->Axis[i].CurVelocity = 0;
					}

					brd->mach_stat = PROBE_SW_HIT;

					tmp = brd->mach_cfg;
					tmp |= PROBE_SW_EN;
					brd->mach_cfg = tmp;
					dummy = brd->mach_cfg;

					brd->irq_raw = CP_IRQ_MACH_PROBE;
					dummy = brd->irq_raw;

					FdoData->IRQMask |= CP_IRQ_MACH_PROBE;

					break;
				}

				if (FdoData->DrvState == DrvJogging) {

					mach->Busy = TRUE;
					mach->Jogging = TRUE;
					for (i = 0; i < 6; i++) {
						AxData[i].PendingIDX = mach->Axis[i].Index;
						AxData[i].Active = FALSE;
						AxData[i].Hit = FALSE;
						AxData[i].QuantCntr = 0;
						mach->Axis[i].CurVelocity = 0;
					}
					RESTORE_LIMITS
					break;
				}

				//
				// Nothing to do, go Idle...
				//
				if (FdoData->DrvState != DrvIdle) {
					FdoData->DrvState = DrvIdle;
					for (i = 0; i < 6; i++) {
						AxData[i].Active = FALSE;
						AxData[i].Hit = FALSE;
						AxData[i].QuantCntr = 0;
						mach->Axis[i].CurVelocity = 0;
					}
				}
				FdoData->StateChange = TRUE;
				mach->Jogging = FALSE;
				mach->Busy = FALSE;
			}
		}
	} while (FALSE);

#ifdef DEBUG
	if (FdoData->DrvState != FdoData->DrvPrevState) {
		DebugPrint(TRACE, DBG_INTERRUPT, "*******************************\n");
		DebugPrint(TRACE, DBG_INTERRUPT, "%s ---> %s%s\n",
			DbgDrvFSMStateString(FdoData->DrvPrevState),
				DbgDrvFSMStateString(FdoData->DrvState),
				mach->Threading ? "(THREADING)" : "");
		DebugPrint(TRACE, DBG_INTERRUPT, "*******************************\n");
		FdoData->DrvPrevState = FdoData->DrvState;
	}
#endif

	//
	// Acknowledge and reenable interrupt(s)
	//
	brd->irq_raw = IrqStatus;
	dummy = brd->irq_raw;

#ifdef _ISR_USE_DPC_
	IoRequestDpc(FdoData->Self, NULL, FdoData);
#else

	brd->irq_mask = FdoData->IRQMask;
	dummy = brd->irq_mask;
#endif

	return TRUE;
}


#ifdef _ISR_USE_DPC_
// ************************
// * DPC Callback for ISR *
// ************************
VOID CPDDpcForIsr(PKDPC Dpc, PDEVICE_OBJECT DeviceObject, PIRP Irp, PVOID Context)
{
	PFDO_DATA	FdoData = (PFDO_DATA)Context;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_DPC, "--> CPDDpcForIsr\n");
#endif

	//
	// Re-enable the interrupt (disabled in ISR)
	//
	KeSynchronizeExecution(FdoData->Interrupt,
					(PKSYNCHRONIZE_ROUTINE)CPDEnableInterrupts, FdoData);

#ifdef DEBUG
	DebugPrint(TRACE, DBG_DPC, "<-- CPDDpcForIsr\n");
#endif

}

VOID CPDEnableInterrupts(PFDO_DATA FdoData)
{
	UINT32	dummy;

	FdoData->CPDRegsBase->irq_mask = FdoData->IRQMask;
	dummy = FdoData->CPDRegsBase->irq_mask;
}
#endif
